# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation nor
# does it submit to any jurisdiction.
#


class PandasHelper:
    def __init__(self, frame, margins=0, **kwargs):

        self.frame = frame
        self.kwargs = kwargs
        self.margins = margins

        if "lat@hdr" in self.frame:
            self.lat = "lat@hdr"
            self.lon = "lon@hdr"
        elif "lat" in self.frame:
            self.lat = "lat"
            self.lon = "lon"
        else:
            self.lat = "latitude"
            self.lon = "longitude"

    def plot_map(self, driver):

        north, east = self.frame[[self.lat, self.lon]].max()
        south, west = self.frame[[self.lat, self.lon]].min()

        driver.bounding_box(
            north=north + self.margins,
            south=south - self.margins,
            west=west - self.margins,
            east=east + self.margins,
        )

        path = "tmp.geo"
        # driver.plot_grib(self.path, self.handle.get('offset'))
        with open(path, "w") as f:
            print("#GEO", file=f)
            print("#lat long value", file=f)
            print("#DATA", file=f)

            seen = set()

            for index, row in self.frame[[self.lat, self.lon]].iterrows():
                if (row[0], row[1]) not in seen:
                    print(row[0], row[1], 42.0, file=f)
                    seen.add((row[0], row[1]))

            # print("----", len(seen))

        driver.plot_geopoints(path)
        driver.apply_kwargs(self.kwargs)

    def bounding_box(self):

        north, east = self.frame[[self.lat, self.lon]].max()
        south, west = self.frame[[self.lat, self.lon]].min()

        return [
            north + self.margins,
            west - self.margins,
            south - self.margins,
            east + self.margins,
        ]

    def dates(self):
        return sorted(set([str(x).split("T")[0] for x in self.frame["time"].values]))


helper = PandasHelper



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation nor
# does it submit to any jurisdiction.
#

import cdsapi
import os

from .base import FileSource, APIKeyPrompt
from climetlab.core.caching import temp_file
from climetlab.helpers import helper


APIRC = "key: {key}\nurl: https://cds.climate.copernicus.eu/api/v2"

MESSAGE = """
An API key is needed to access this dataset. Please visit
<https://cds.climate.copernicus.eu/> to register or sign-in,
and visit <https://cds.climate.copernicus.eu/api-how-to> to
retrieve you API key.

Once this is done, please paste your key in the input field below
and press *ENTER*.
"""


class CDSAPI(APIKeyPrompt):

    text_message = MESSAGE

    markdown_message = MESSAGE

    rcfile = "~/.cdsapirc"
    prompt = "CDS api key"

    def validate(self, text):
        uid, key = text.strip().split(":")
        return APIRC.format(key="%s:%s" % (uid, key))


def client():
    try:
        return cdsapi.Client()
    except Exception as e:
        if ".cdsapirc" in str(e):
            CDSAPI().ask_user_and_save()
            return cdsapi.Client()
        else:
            raise


class CDSRetriever(FileSource):

    sphinxdoc = """
    CDSRetriever
    """

    def __init__(self, dataset, **kwargs):
        request = self.request(**kwargs)
        self.path = temp_file("CDSRetriever", request)
        if not os.path.exists(self.path):
            client().retrieve(dataset, request, self.path + ".tmp")
            os.rename(self.path + ".tmp", self.path)

    def request(self, **kwargs):
        based_on = kwargs.pop("based_on", None)
        if based_on is not None:
            data = based_on.pop("data")
            data = helper(data, **based_on)
            kwargs["area"] = data.bounding_box()
            kwargs["date"] = data.dates()

        return kwargs

    @property
    def read_csv_options(self):
        return dict(
            comment="#",
            parse_dates=["report_timestamp"],
            skip_blank_lines=True,
            compression="zip",
        )


source = CDSRetriever



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation nor
# does it submit to any jurisdiction.
#

from importlib import import_module
import os
import weakref


def lookup(name):
    source = import_module(".%s" % (name.replace("-", "_"),), package=__name__)
    return source.source


def load(name, *args, **kwargs):
    source = lookup(name)(*args, **kwargs)
    if source.name is None:
        source.name = name
    return source


def list_entries():
    here = os.path.realpath(os.path.dirname(__file__))
    result = []

    for n in os.listdir(here):
        if n.startswith("."):
            continue

        if n.startswith("_"):
            continue

        if not n.endswith(".py"):
            continue

        result.append(n[:-3])

    return result


class DataSource:

    name = None
    home_page = "-"
    licence = "-"
    documentation = "-"

    _dataset = None

    def __init__(self, **kwargs):
        self._kwargs = kwargs

    @property
    def dataset(self):
        if self._dataset is None:
            return None
        return self._dataset()

    @dataset.setter
    def dataset(self, dataset):
        self._dataset = weakref.ref(dataset)

    def _repr_markdown_(self):
        return """

| {name} |  |
| --- | --- |
| Home page | {home_page} |
| Documentation | {documentation} |
| Licence | {licence} |
        """.format(
            name=self.name,
            home_page=self.home_page,
            licence=self.licence,
            documentation=self.documentation,
        )

    # def __iter__(self):
    #     raise NotImplementedError("%r: method __iter__() not implemented" % (self.__class__.__name__,))

    # def __len__(self):
    #     raise NotImplementedError("%r: method __len__() not implemented" % (self.__class__.__name__,))

    # def __getitem__(self, n):
    #     raise NotImplementedError("%r: method __getitem__() not implemented" % (self.__class__.__name__,))

    # def to_xarray(self, *args, **kwargs):
    #     raise NotImplementedError("%r: method to_xarray() not implemented" % (self.__class__.__name__,))

    # def to_pandas(self, *args, **kwargs):
    #     raise NotImplementedError("%r: method to_pandas() not implemented" % (self.__class__.__name__,))

    # def to_numpy(self, *args, **kwargs):
    #     raise NotImplementedError("%r: method to_numpy() not implemented" % (self.__class__.__name__,))

    # def to_metview(self, *args, **kwargs):
    #     raise NotImplementedError("%r: method to_metview() not implemented" % (self.__class__.__name__,))



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation nor
# does it submit to any jurisdiction.
#

import numpy as np
import climetlab
from importlib import import_module
from climetlab.core.plugins import find_plugin
import os
import yaml


class Dataset:

    name = None
    home_page = "-"
    licence = "-"
    documentation = "-"
    citation = "-"

    _source = None

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, source):
        self._source = source
        source.dataset = self

    def __len__(self):
        return len(self.source)

    def __getitem__(self, n):
        return self.source[n]

    def to_numpy(self, *args, **kwargs):
        return np.array([s.to_numpy(*args, **kwargs) for s in self.source])

    def to_xarray(self, *args, **kwargs):
        return self.source.to_xarray(*args, **kwargs)

    def to_pandas(self, *args, **kwargs):
        return self.source.to_pandas(*args, **kwargs)

    def to_metview(self, *args, **kwargs):
        return self.source.to_metview(*args, **kwargs)

    def _repr_markdown_(self):
        return """

| {name} |  |
| --- | --- |
| Home page | {home_page} |
| Documentation | {documentation} |
| Citation | {citation} |
| Licence | {licence} |
        """.format(
            name=self.name,
            home_page=self.home_page,
            licence=self.licence,
            citation="<pre>%s</pre>" % (self.citation.replace("\n", "<br>"),),
            documentation=self.documentation,
        )


def _module_callback(plugin):
    return import_module(plugin, package=__name__).dataset


def _yaml_callcack(plugin):
    with open(plugin) as f:
        dataset = yaml.load(f.read(), Loader=yaml.SafeLoader)["dataset"]

    class Wrapped(Dataset):
        def __init__(self, *args, **kwargs):
            self.source = climetlab.load_source(
                dataset["source"], **dataset.get("args", {})
            )

    return Wrapped


def _lookup(name):
    return find_plugin(
        os.path.dirname(__file__),
        name,
        module_callback=_module_callback,
        yaml_callcack=_yaml_callcack,
    )


def load(name, *args, **kwargs):
    dataset = _lookup(name)(*args, **kwargs)
    if dataset.name is None:
        dataset.name = name
    return dataset



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation nor
# does it submit to any jurisdiction.
#

from .bindings import (
    grib_file_open,
    grib_handle_delete,
    grib_get,
    grib_values,
    GribError,
)


class GribField:
    def __init__(self, handle, path, offset):
        self._delete = grib_handle_delete
        self.handle = handle
        self.path = path
        self.offset = offset

    def __del__(self):
        self._delete(self.handle)

    def get(self, name):
        try:
            if name == "values":
                return grib_values(self.handle)
            return grib_get(self.handle, name)
        except GribError as e:
            if e.err == -10:  # Key not found
                return None
            raise


class GribFile:
    def __init__(self, path):
        self.path = path
        self.file = grib_file_open(path)

    def __del__(self):
        try:
            self.file.close()
        except Exception:
            pass

    def __iter__(self):
        return self

    def __next__(self):
        here = self.file.tell()
        h = self.file.next()
        if not h:
            raise StopIteration()
        return GribField(h, self.path, here)

    def at_offset(self, offset):
        self.file.position(offset)
        return self.__next__()



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation nor
# does it submit to any jurisdiction.
#

from .era5_single_levels import Era5SingleLevels


class Era5Temperature(Era5SingleLevels):
    def __init__(self, *args, **kargs):
        super().__init__("2t", *args, **kargs)


dataset = Era5Temperature



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation nor
# does it submit to any jurisdiction.
#

import os


CACHE = {}


def _find_plugin(directory, name):
    n = len(directory)
    for path, dirs, files in os.walk(directory):
        path = path[n:]
        for f in files:
            if f.endswith(".yaml"):
                if f[:-5] == name:
                    return False, os.path.join(directory, path, f)

            if f.endswith(".py"):
                p = os.path.join(path, f[:-3])
                if p[0] != "/":
                    p = "/" + p
                if p[1:].replace("/", "-").replace("_", "-") == name:
                    return True, p.replace("/", ".")

    raise Exception("Cannot find plugin '%s' in %s" % (name, directory))


def find_plugin(directory, name, module_callback, yaml_callcack):
    if (directory, name) not in CACHE:
        CACHE[(directory, name)] = _find_plugin(directory, name)

    code, plugin = CACHE[(directory, name)]
    if code:
        return module_callback(plugin)
    else:
        return yaml_callcack(plugin)
